@@ -22,6 +22,7 @@
 */
 
 #include <config.h>
+#include <math.h>
 #include <string.h>
 #include <stdio.h>
 #include "tasklist.h"
@@ -62,6 +63,8 @@ typedef struct _WnckTaskClass   WnckTaskClass;
 #define DEFAULT_WIDTH 1
 #define DEFAULT_HEIGHT 48
 
+#define MAX_WIDTH_CHARS 250
+
 #define N_SCREEN_CONNECTIONS 5
 
 #define POINT_IN_RECT(xcoord, ycoord, rect) \
@@ -116,6 +119,14 @@ struct _WnckTask
   guint was_active : 1;      /* used to fixup activation behavior */ 
 
   guint button_activate;
+
+  guint32 dnd_timestamp;
+
+  GdkPixbuf *screenshot;
+
+  gdouble glow_start_time;
+  
+  guint button_glow;
 };
 
 struct _WnckTaskClass
@@ -174,6 +185,7 @@ struct _WnckTasklistPrivate
 
   gint monitor_num;
   GdkRectangle monitor_geometry;
+  GtkReliefStyle relief;
 };
 
 
@@ -189,8 +201,10 @@ static WnckTask *wnck_task_new_from_class_group (WnckTasklist    *tasklist,
 static WnckTask *wnck_task_new_from_startup_sequence (WnckTasklist      *tasklist,
                                                       SnStartupSequence *sequence);
 #endif
+static gboolean wnck_task_get_needs_attention (WnckTask *task);
 
-static char      *wnck_task_get_text (WnckTask *task);
+
+static char      *wnck_task_get_text (WnckTask *task, gboolean icon_text);
 static GdkPixbuf *wnck_task_get_icon (WnckTask *task);
 static gint       wnck_task_compare  (gconstpointer  a,
 				      gconstpointer  b);
@@ -240,7 +254,8 @@ static void     wnck_tasklist_connect_window           (WnckTasklist *tasklist,
 static void     wnck_tasklist_change_active_task       (WnckTasklist *tasklist,
 							WnckTask *active_task);
 static gboolean wnck_tasklist_change_active_timeout    (gpointer data);
-static void     wnck_tasklist_activate_task_window     (WnckTask *task);
+static void     wnck_tasklist_activate_task_window     (WnckTask *task,
+                                                        guint32   timestamp);
 
 static void     wnck_tasklist_update_icon_geometries   (WnckTasklist *tasklist,
 							GList        *visible_tasks);
@@ -312,7 +327,10 @@ make_gtk_label_normal (GtkLabel *label)
   pango_font_description_free (font_desc);
 }
 
-GType
+static GType
+wnck_task_get_type (void) G_GNUC_CONST;
+
+static GType
 wnck_task_get_type (void)
 {
   static GType object_type = 0;
@@ -332,6 +350,7 @@ wnck_task_get_type (void)
         sizeof (WnckTask),
         0,              /* n_preallocs */
         (GInstanceInitFunc) wnck_task_init,
+        NULL            /* value_table */
       };
       
       object_type = g_type_register_static (G_TYPE_OBJECT,
@@ -345,6 +364,9 @@ wnck_task_get_type (void)
 static void
 wnck_task_init (WnckTask *task)
 {  
+  task->glow_start_time = 0.0;
+  task->button_glow = 0;
+  task->screenshot = NULL;
 }
 
 static void
@@ -366,6 +388,147 @@ wnck_task_class_init (WnckTaskClass *klass)
     "    widget \"*.tasklist-button\" style \"tasklist-button-style\"\n"
     "\n");
 }
+static GdkPixbuf *
+glow_pixbuf (const GdkPixbuf *source, 
+             gdouble          factor)
+{
+  gint i, j;
+  gint width, height, has_alpha, rowstride;
+  guchar *destination_data;
+  guchar *source_data;
+  guchar *source_pixel;
+  guchar *destination_pixel;
+  int channel_intensity;
+  guchar r, g, b;
+
+  GdkPixbuf *destination;
+
+  has_alpha = gdk_pixbuf_get_has_alpha (source);
+  width = gdk_pixbuf_get_width (source);
+  height = gdk_pixbuf_get_height (source);
+  rowstride = gdk_pixbuf_get_rowstride (source);
+  source_data = gdk_pixbuf_get_pixels (source);
+
+  destination = gdk_pixbuf_copy (source);
+  destination_data = gdk_pixbuf_get_pixels (destination);
+
+  for (i = 0; i < height; i++)
+    {
+      source_pixel = source_data + i * rowstride;
+      destination_pixel = destination_data + i * rowstride;
+
+      for (j = 0; j < width; j++)
+        {
+          enum
+            {
+              CHANNEL_RED = 0,
+              CHANNEL_GREEN,
+              CHANNEL_BLUE,
+              CHANNELS_PER_COLOR_STIMULUS
+            };
+          r = source_pixel[CHANNEL_RED];
+          g = source_pixel[CHANNEL_GREEN];
+          b = source_pixel[CHANNEL_BLUE];
+          source_pixel += CHANNELS_PER_COLOR_STIMULUS;
+
+          channel_intensity = r * factor;
+          destination_pixel[CHANNEL_RED] = CLAMP (channel_intensity, 0, 255);
+          channel_intensity = g * factor;
+          destination_pixel[CHANNEL_GREEN] = CLAMP (channel_intensity, 0, 255);
+          channel_intensity = b * factor;
+          destination_pixel[CHANNEL_BLUE] = CLAMP (channel_intensity, 0, 255);
+          destination_pixel += CHANNELS_PER_COLOR_STIMULUS;
+
+          if (has_alpha)
+            {
+              destination_pixel[0] = source_pixel[0];
+              source_pixel++;
+              destination_pixel++;
+            }
+        }
+    }
+  return destination;
+}
+
+static gboolean
+wnck_task_button_glow (WnckTask *task)
+{
+  GdkPixbuf *glowing_screenshot;
+  GTimeVal tv;
+  gdouble glow_factor, now;
+
+  if (task->screenshot == NULL)
+    return TRUE;
+
+  g_get_current_time (&tv);
+  now = (tv.tv_sec * (1.0 * G_USEC_PER_SEC) +
+        tv.tv_usec) / G_USEC_PER_SEC;
+
+  if (task->glow_start_time <= G_MINDOUBLE)
+    task->glow_start_time = now;
+
+  /* These numbers can probably be tweaked some.
+   * The 1.0 is the first value it will get and .4 is
+   * the "radius" from 1.0 that it will go; i.e.,
+   * it will go up from 1.0 to 1.4 then down from
+   * 1.4 to .6, then back up again. The 2.0 in the
+   * denomator is how many seconds it takes to travel 
+   * through the entire range of numbers.
+   */
+  glow_factor = .4 * sin ((now - task->glow_start_time) * (2.0 * M_PI) / 2.0) + 1.0;
+
+  glowing_screenshot = glow_pixbuf (task->screenshot, glow_factor);
+
+  gdk_draw_pixbuf (task->button->window,
+                   task->button->style->fg_gc[GTK_WIDGET_STATE (task->button)],
+                   glowing_screenshot,
+                   0, 0, 
+                   task->button->allocation.x, 
+                   task->button->allocation.y,
+                   gdk_pixbuf_get_width (glowing_screenshot),
+                   gdk_pixbuf_get_height (glowing_screenshot),
+                   GDK_RGB_DITHER_NORMAL, 0, 0);
+  gdk_pixbuf_unref (glowing_screenshot);
+
+  return TRUE;
+}
+
+static void
+wnck_task_clear_glow_start_timeout_id (WnckTask *task)
+{
+  task->button_glow = 0;
+}
+
+static void
+wnck_task_queue_glow (WnckTask *task)
+{
+  if (task->button_glow == 0)
+    {
+      task->glow_start_time = 0.0;
+
+      /* The animation doesn't speed up or slow down based on the
+       * timeout value, but instead will just appear smoother or 
+       * choppier. To adjust the speed change the parameter to 
+       * sine in the timeout callback.
+       */
+      task->button_glow =
+        g_timeout_add_full (G_PRIORITY_DEFAULT_IDLE, 
+                            50,
+                            (GSourceFunc) wnck_task_button_glow, task,
+                            (GDestroyNotify) wnck_task_clear_glow_start_timeout_id);
+    }
+}
+
+static void
+wnck_task_stop_glow (WnckTask *task)
+{
+  if (task->button_glow != 0)
+    {
+      g_source_remove (task->button_glow);
+      task->button_glow = 0;
+      task->glow_start_time = 0.0;
+    }
+}
 
 static void
 wnck_task_finalize (GObject *object)
@@ -459,6 +622,8 @@ wnck_task_finalize (GObject *object)
       task->button_activate = 0;
     } 
 
+  wnck_task_stop_glow (task);
+
   G_OBJECT_CLASS (task_parent_class)->finalize (object);
 }
 
@@ -482,6 +647,7 @@ wnck_tasklist_get_type (void)
         sizeof (WnckTasklist),
         0,              /* n_preallocs */
         (GInstanceInitFunc) wnck_tasklist_init,
+        NULL            /* value_table */
       };
       
       object_type = g_type_register_static (GTK_TYPE_CONTAINER,
@@ -518,6 +684,7 @@ wnck_tasklist_init (WnckTasklist *tasklist)
   tasklist->priv->idle_callback_tag = 0;
 
   tasklist->priv->monitor_num = -1;
+  tasklist->priv->relief = GTK_RELIEF_NORMAL;
 
   atk_obj = gtk_widget_get_accessible (widget);
   atk_object_set_name (atk_obj, _("Window List"));
@@ -615,6 +782,27 @@ wnck_tasklist_set_grouping (WnckTasklist            *tasklist,
   gtk_widget_queue_resize (GTK_WIDGET (tasklist));
 }
 
+static void
+wnck_tasklist_set_relief_callback (WnckWindow   *win,
+				   WnckTask     *task,
+				   WnckTasklist *tasklist)
+{
+  gtk_button_set_relief (GTK_BUTTON (task->button), tasklist->priv->relief);
+}
+
+void
+wnck_tasklist_set_button_relief (WnckTasklist *tasklist, GtkReliefStyle relief)
+{
+  if (relief == tasklist->priv->relief)
+    return;
+
+  tasklist->priv->relief = relief;
+
+  g_hash_table_foreach (tasklist->priv->win_hash,
+                        (GHFunc) wnck_tasklist_set_relief_callback,
+                        tasklist);
+}
+
 void
 wnck_tasklist_set_switch_workspace_on_unminimize (WnckTasklist  *tasklist,
 						  gboolean       switch_workspace_on_unminimize)
@@ -1112,6 +1300,12 @@ wnck_tasklist_size_allocate (GtkWidget      *widget,
 	      win_task = WNCK_TASK (l->data);
 	      
 	      gtk_widget_set_child_visible (GTK_WIDGET (win_task->button), FALSE);
+
+              if (win_task->screenshot != NULL)
+                {
+                  g_object_unref (win_task->screenshot);
+                  win_task->screenshot = NULL;
+                }
 	      l = l->next;
 	    }
 	}
@@ -1119,7 +1313,13 @@ wnck_tasklist_size_allocate (GtkWidget      *widget,
 	{
 	  visible_tasks = g_list_prepend (visible_tasks, class_group_task->windows->data);
 	  gtk_widget_set_child_visible (GTK_WIDGET (class_group_task->button), FALSE);
-	}
+
+          if (class_group_task->screenshot != NULL)
+            {
+              g_object_unref (class_group_task->screenshot);
+              class_group_task->screenshot = NULL;
+            }
+        }
       
       button_width = wnck_tasklist_layout (allocation,
 					   tasklist->priv->max_button_width,
@@ -1136,6 +1336,12 @@ wnck_tasklist_size_allocate (GtkWidget      *widget,
       
       visible_tasks = g_list_concat (visible_tasks, g_list_copy (class_group_task->windows));
       gtk_widget_set_child_visible (GTK_WIDGET (class_group_task->button), FALSE);
+      if (class_group_task->screenshot != NULL)
+        {
+          g_object_unref (class_group_task->screenshot);
+          class_group_task->screenshot = NULL;
+        }
+
       l = l->next;
     }
 
@@ -1557,7 +1763,7 @@ wnck_tasklist_update_lists (WnckTasklist *tasklist)
 	  
 	  gtk_widget_set_parent (win_task->button, GTK_WIDGET (tasklist));
 	  gtk_widget_show (win_task->button);
-	  
+
 	  /* Class group */
 
 	  class_group = wnck_window_get_class_group (win);
@@ -1701,12 +1907,14 @@ wnck_tasklist_active_window_changed (WnckScreen   *screen,
 				     WnckTasklist *tasklist)
 {
   WnckWindow *active_window;
-  WnckTask *active_task;
-  
-  active_window = wnck_screen_get_active_window (screen);
+  WnckTask *active_task = NULL;
 
-  active_task = g_hash_table_lookup (tasklist->priv->win_hash,
-				     active_window);
+  /* FIXME: check for group modal window */
+  for (active_window = wnck_screen_get_active_window (screen);
+       active_window && !active_task;
+       active_window = wnck_window_get_transient(active_window))
+    active_task = g_hash_table_lookup (tasklist->priv->win_hash,
+                                       active_window);
 
   wnck_tasklist_change_active_task (tasklist, active_task);
 }
@@ -1915,11 +2123,15 @@ wnck_task_menu_activated (GtkMenuItem *menu_item,
 {
   WnckTask *task = WNCK_TASK (data);
 
-  wnck_tasklist_activate_task_window (task);
+  /* This is an "activate" callback function so gtk_get_current_event_time()
+   * will suffice.
+   */
+  wnck_tasklist_activate_task_window (task, gtk_get_current_event_time ());
 }
 
 static void
-wnck_tasklist_activate_task_window (WnckTask *task)
+wnck_tasklist_activate_task_window (WnckTask *task,
+                                    guint32   timestamp)
 {
   WnckTasklist *tasklist;
   WnckWindowState state;
@@ -1940,14 +2152,15 @@ wnck_tasklist_activate_task_window (WnckTask *task)
       window_ws = wnck_window_get_workspace (task->window);
       if (window_ws &&
           active_ws != window_ws &&
-	  !tasklist->priv->switch_workspace_on_unminimize)
-	wnck_workspace_activate (window_ws);
-	  
-      wnck_window_activate_transient (task->window);
+          !tasklist->priv->switch_workspace_on_unminimize)
+        wnck_workspace_activate (window_ws, timestamp);
+
+      wnck_window_activate_transient (task->window, timestamp);
     }
   else
     {
-      if (task->was_active || wnck_window_transient_is_active (task->window))
+      if (task->was_active || 
+          wnck_window_transient_is_most_recently_activated (task->window))
 	{
 	  task->was_active = FALSE;
 	  wnck_window_minimize (task->window);
@@ -1957,12 +2170,16 @@ wnck_tasklist_activate_task_window (WnckTask *task)
 	{
           WnckWorkspace *window_ws;
           
+          /* FIXME: THIS IS SICK AND WRONG AND BUGGY.  See the end of
+           * http://mail.gnome.org/archives/wm-spec-list/2005-July/msg00032.html
+           * There should only be *one* activate call.
+           */
           window_ws = wnck_window_get_workspace (task->window);
           if (window_ws)
-            wnck_workspace_activate (window_ws);
+            wnck_workspace_activate (window_ws, timestamp);
 
-	  wnck_window_activate_transient (task->window);
-	}
+          wnck_window_activate_transient (task->window, timestamp);
+        }
     }
   
 
@@ -2002,7 +2219,10 @@ wnck_task_unminimize_all (GtkMenuItem *menu_item,
   while (l)
     {
       WnckTask *child = WNCK_TASK (l->data);
-      wnck_window_unminimize (child->window);
+      /* This is inside an activate callback, so gtk_get_current_event_time()
+       * will work.
+       */
+      wnck_window_unminimize (child->window, gtk_get_current_event_time ());
       l = l->next;
     }
 }
@@ -2059,8 +2279,10 @@ wnck_task_popup_menu (WnckTask *task,
     {
       win_task = WNCK_TASK (l->data);
       
-      text = wnck_task_get_text (win_task);
+      text = wnck_task_get_text (win_task, FALSE);
       menu_item = gtk_image_menu_item_new_with_label (text);
+      if (wnck_task_get_needs_attention (win_task)) 
+        make_gtk_label_bold (GTK_LABEL (GTK_BIN (menu_item)->child));
       g_free (text);
       
       pixbuf = wnck_task_get_icon (win_task);
@@ -2165,8 +2387,11 @@ wnck_task_button_toggled (GtkButton *button,
     case WNCK_TASK_WINDOW:
       if (task->window == NULL)
 	return;
-      
-      wnck_tasklist_activate_task_window (task);
+
+      /* This should only be called by clicking on the task button, so
+       * gtk_get_current_event_time() should be fine here...
+       */
+      wnck_tasklist_activate_task_window (task, gtk_get_current_event_time ());
       break;
     case WNCK_TASK_STARTUP_SEQUENCE:
       break;
@@ -2174,7 +2399,7 @@ wnck_task_button_toggled (GtkButton *button,
 }
 
 static char *
-wnck_task_get_text (WnckTask *task)
+wnck_task_get_text (WnckTask *task, gboolean icon_text)
 {
   WnckWindowState state;
   const char *name;
@@ -2192,8 +2417,12 @@ wnck_task_get_text (WnckTask *task)
 				g_list_length (task->windows));
 
     case WNCK_TASK_WINDOW:
+      if (icon_text)
+        name = wnck_window_get_icon_name (task->window);
+      else 
+        name = wnck_window_get_name (task->window);
+       
       state = wnck_window_get_state (task->window);
-      name = wnck_window_get_name (task->window);
       
       if (state & WNCK_WINDOW_STATE_SHADED)
 	return g_strdup_printf ("=%s=", name);
@@ -2360,13 +2589,13 @@ wnck_task_get_icon (WnckTask *task)
 }
 
 static gboolean
-wnck_task_get_demands_attention (WnckTask *task)
+wnck_task_get_needs_attention (WnckTask *task)
 {
   GList *l;
   WnckTask *win_task;
-  gboolean demands_attention;
+  gboolean needs_attention;
 
-  demands_attention = FALSE;
+  needs_attention = FALSE;
 
   switch (task->type)
     {
@@ -2376,9 +2605,9 @@ wnck_task_get_demands_attention (WnckTask *task)
 	{
 	  win_task = WNCK_TASK (l->data);
 
-	  if (wnck_window_get_state (win_task->window) & WNCK_WINDOW_STATE_DEMANDS_ATTENTION)
+	  if (wnck_window_or_transient_needs_attention (win_task->window))
 	    {
-	      demands_attention = TRUE;
+	      needs_attention = TRUE;
 	      break;
 	    }
 
@@ -2387,15 +2616,15 @@ wnck_task_get_demands_attention (WnckTask *task)
       break;
 
     case WNCK_TASK_WINDOW:
-      demands_attention =
-	wnck_window_get_state (task->window) & WNCK_WINDOW_STATE_DEMANDS_ATTENTION;
+      needs_attention =
+	wnck_window_or_transient_needs_attention (task->window);
       break;
 
     case WNCK_TASK_STARTUP_SEQUENCE:
       break;
     }
 
-  return demands_attention != FALSE;
+  return needs_attention != FALSE;
 }
 
 static void
@@ -2410,16 +2639,27 @@ wnck_task_update_visible_state (WnckTask *task)
   if (pixbuf)
     g_object_unref (pixbuf);
 
-  text = wnck_task_get_text (task);
+  text = wnck_task_get_text (task, TRUE);
   if (text != NULL)
     {
       gtk_label_set_text (GTK_LABEL (task->label), text);
-      if (wnck_task_get_demands_attention (task))
-        make_gtk_label_bold ((GTK_LABEL (task->label)));
+      if (wnck_task_get_needs_attention (task))
+        {
+          make_gtk_label_bold ((GTK_LABEL (task->label)));
+          wnck_task_queue_glow (task);
+        }
       else
-        make_gtk_label_normal ((GTK_LABEL (task->label)));
-      gtk_tooltips_set_tip (task->tasklist->priv->tooltips, task->button, text, NULL);
+        {
+          make_gtk_label_normal ((GTK_LABEL (task->label)));
+          wnck_task_stop_glow (task);
+        }
       g_free (text);
+      text = wnck_task_get_text (task, FALSE);
+      if (text != NULL)
+        {
+          gtk_tooltips_set_tip (task->tasklist->priv->tooltips, task->button, text, NULL);
+          g_free (text);
+        }
     }
 
   gtk_widget_queue_resize (GTK_WIDGET (task->tasklist));
@@ -2440,8 +2680,9 @@ wnck_task_state_changed (WnckWindow     *window,
       return;
     }
   
-  if ((changed_mask & WNCK_WINDOW_STATE_MINIMIZED) ||
-      (changed_mask & WNCK_WINDOW_STATE_DEMANDS_ATTENTION))
+  if ((changed_mask & WNCK_WINDOW_STATE_MINIMIZED)         ||
+      (changed_mask & WNCK_WINDOW_STATE_DEMANDS_ATTENTION) ||
+      (changed_mask & WNCK_WINDOW_STATE_URGENT))
     {
       WnckTask *win_task;
 
@@ -2510,7 +2751,9 @@ wnck_task_motion_timeout (gpointer data)
 
   task->button_activate = 0;
 
-  wnck_window_activate_transient (task->window);
+  wnck_window_activate_transient (task->window, task->dnd_timestamp);
+
+  task->dnd_timestamp = 0;
 
   return FALSE;
 }
@@ -2536,14 +2779,14 @@ wnck_task_drag_motion (GtkWidget          *widget,
 		       guint               time,
 		       WnckTask            *task)
 {
+  task->dnd_timestamp = time;
+  if (task->button_activate == 0 && task->type == WNCK_TASK_WINDOW)
+      task->button_activate = g_timeout_add (WNCK_ACTIVATE_TIMEOUT,
+                                             wnck_task_motion_timeout,
+                                             task);
+  gdk_drag_status (context,0,time);
 
-   if (task->button_activate == 0 && task->type == WNCK_TASK_WINDOW)
-       task->button_activate = g_timeout_add (WNCK_ACTIVATE_TIMEOUT,
-                                              wnck_task_motion_timeout,
-                                              task);
-   gdk_drag_status (context,0,time);
-
-   return TRUE;
+  return TRUE;
 }
 
 static gboolean
@@ -2611,11 +2854,15 @@ wnck_task_button_press_event (GtkWidget	      *widget,
   return FALSE;
 }
 
+static gboolean
+wnck_task_expose (GtkWidget        *widget,
+                  GdkEventExpose   *event,
+                  gpointer          data);
 
 static void
-wnck_task_create_widgets (WnckTask *task)
+wnck_task_create_widgets (WnckTask *task, GtkReliefStyle relief)
 {
-  GtkWidget *table;
+  GtkWidget *hbox;
   GdkPixbuf *pixbuf;
   char *text;
   static GQuark disable_sound_quark = 0;
@@ -2627,6 +2874,9 @@ wnck_task_create_widgets (WnckTask *task)
     task->button = gtk_button_new ();
   else
     task->button = gtk_toggle_button_new ();
+
+  gtk_button_set_relief (GTK_BUTTON (task->button), relief);
+
   task->button_activate = 0;
   g_object_set_qdata (G_OBJECT (task->button),
 		      disable_sound_quark, GINT_TO_POINTER (TRUE));
@@ -2638,7 +2888,7 @@ wnck_task_create_widgets (WnckTask *task)
 
   gtk_drag_dest_set (GTK_WIDGET(task->button), 0, NULL, 0, 0);
 
-  table = gtk_table_new (1, 2, FALSE);
+  hbox = gtk_hbox_new (FALSE, 0);
 
   pixbuf = wnck_task_get_icon (task);
   if (pixbuf)
@@ -2651,32 +2901,29 @@ wnck_task_create_widgets (WnckTask *task)
   
   gtk_widget_show (task->image);
 
-  text = wnck_task_get_text (task);
+  text = wnck_task_get_text (task, TRUE);
   task->label = gtk_label_new (text);
   gtk_misc_set_alignment (GTK_MISC (task->label), 0.0, 0.5);
   gtk_label_set_ellipsize (GTK_LABEL (task->label),
                           PANGO_ELLIPSIZE_END);
+  gtk_label_set_max_width_chars (GTK_LABEL (task->label), MAX_WIDTH_CHARS);
+
+  if (wnck_task_get_needs_attention (task))
+    {
+      make_gtk_label_bold ((GTK_LABEL (task->label)));
+      wnck_task_queue_glow (task);
+    }
 
-  if (wnck_task_get_demands_attention (task))
-    make_gtk_label_bold ((GTK_LABEL (task->label)));
   gtk_widget_show (task->label);
 
-  gtk_table_attach (GTK_TABLE (table),
-		    task->image,
-		    0, 1,
-		    0, 1,
-		    0, GTK_EXPAND,
-		    4, 0);
-  gtk_table_attach (GTK_TABLE (table),
-		    task->label,
-		    1, 2,
-		    0, 1,
-		    GTK_FILL | GTK_EXPAND, GTK_EXPAND,		    
-		    2, 0);
+  gtk_box_pack_start (GTK_BOX (hbox), task->image, FALSE, FALSE, 4);
+  gtk_box_pack_start (GTK_BOX (hbox), task->label, TRUE, TRUE, 2);
 
-  gtk_container_add (GTK_CONTAINER (task->button), table);
-  gtk_widget_show (table);
+  gtk_container_add (GTK_CONTAINER (task->button), hbox);
+  gtk_widget_show (hbox);
+  g_free (text);
   
+  text = wnck_task_get_text (task, FALSE);
   gtk_tooltips_set_tip (task->tasklist->priv->tooltips, task->button, text, NULL);
   g_free (text);
   
@@ -2727,6 +2974,11 @@ wnck_task_create_widgets (WnckTask *task)
     default:
       g_assert_not_reached ();
     }
+
+  g_signal_connect_object (task->button, "expose_event",
+                           G_CALLBACK (wnck_task_expose),
+                           G_OBJECT (task),
+                           G_CONNECT_AFTER);
 }
 
 static void
@@ -2736,30 +2988,74 @@ draw_dot (GdkWindow *window, GdkGC *lgc, GdkGC *dgc, int x, int y)
   gdk_draw_point (window, lgc, x+1, y+1);
 }
 
-
 static gboolean
-wnck_task_class_group_expose (GtkWidget        *widget,
-			      GdkEventExpose   *event,
-			      gpointer          data)
+wnck_task_expose (GtkWidget        *widget,
+                  GdkEventExpose   *event,
+                  gpointer          data)
 {
   GtkStyle *style;
   GdkGC *lgc, *dgc;
   int x, y, i, j;
+  WnckTask *task;
 
-  style = widget->style;
+  task = WNCK_TASK (data);
   
-  lgc = style->light_gc[GTK_STATE_NORMAL];
-  dgc = style->dark_gc[GTK_STATE_NORMAL];
+  switch (task->type)
+    {
+    case WNCK_TASK_CLASS_GROUP:
+      style = widget->style;
+      lgc = style->light_gc[GTK_STATE_NORMAL];
+      dgc = style->dark_gc[GTK_STATE_NORMAL];
 
-  x = widget->allocation.x + widget->allocation.width -
-    (GTK_CONTAINER (widget)->border_width + style->ythickness + 10);
-  y = widget->allocation.y + style->xthickness + 2;
- 
-  for (i = 0; i < 3; i++) {
-    for (j = i; j < 3; j++) {
-      draw_dot (widget->window, lgc, dgc, x + j*3, y + i*3);
+      x = widget->allocation.x + widget->allocation.width -
+          (GTK_CONTAINER (widget)->border_width + style->ythickness + 10);
+      y = widget->allocation.y + style->xthickness + 2;
+
+      for (i = 0; i < 3; i++) {
+            for (j = i; j < 3; j++) {
+                  draw_dot (widget->window, lgc, dgc, x + j*3, y + i*3);
+            }
+      }
+
+      /* Fall through to get screenshot
+       */
+    case WNCK_TASK_WINDOW:
+      if ((event->area.x <= widget->allocation.x) &&
+          (event->area.y <= widget->allocation.y) &&
+          (event->area.width >= widget->allocation.width) &&
+          (event->area.height >= widget->allocation.height))
+        {
+
+          if (task->screenshot != NULL)
+            {
+              g_object_unref (task->screenshot);
+              task->screenshot = NULL;
+            }
+
+          if (task->button_glow != 0)
+            {
+              task->screenshot = gdk_pixbuf_get_from_drawable (NULL,
+                                                               widget->window,
+                                                               NULL,
+                                                               widget->allocation.x, 
+                                                               widget->allocation.y,
+                                                               0, 0,
+                                                               widget->allocation.width, 
+                                                               widget->allocation.height);
+            }
+        }
+
+    case WNCK_TASK_STARTUP_SEQUENCE:
+      break;
     }
-  }
+
+  if ((task->button_glow == 0) &&
+      (task->screenshot != NULL))
+    {
+      g_object_unref (task->screenshot);
+      task->screenshot = NULL;
+    }
+
   return FALSE;
 }
 
@@ -2871,7 +3167,7 @@ wnck_task_new_from_window (WnckTasklist *tasklist,
   task->class_group = g_object_ref (wnck_window_get_class_group (window));
   task->tasklist = tasklist;
   
-  wnck_task_create_widgets (task);
+  wnck_task_create_widgets (task, tasklist->priv->relief);
 
   remove_startup_sequences_for_window (tasklist, window);
   
@@ -2891,12 +3187,7 @@ wnck_task_new_from_class_group (WnckTasklist   *tasklist,
   task->class_group = g_object_ref (class_group);
   task->tasklist = tasklist;
 
-  wnck_task_create_widgets (task);
-
-  g_signal_connect_object (task->button, "expose_event",
-                           G_CALLBACK (wnck_task_class_group_expose),
-                           G_OBJECT (task),
-                           G_CONNECT_AFTER);
+  wnck_task_create_widgets (task, tasklist->priv->relief);
 
   return task;
 }
@@ -2917,7 +3208,7 @@ wnck_task_new_from_startup_sequence (WnckTasklist      *tasklist,
   sn_startup_sequence_ref (task->startup_sequence);
   task->tasklist = tasklist;
   
-  wnck_task_create_widgets (task);
+  wnck_task_create_widgets (task, tasklist->priv->relief);
 
   return task;
 }
